#include <math.h>
#include <LiquidCrystal.h>

//Assignació pins pantalla LCD
const int rs = 13, en = 12, d4 = 11, d5 = 2, d6 = 3, d7 = 4;
LiquidCrystal lcd(rs, en, d4, d5, d6, d7);

/*En funció de si volem mostrar les dades per la pantalla LCD o enviar-les pel port sèrie posarem la seva respectiva variable a 1, però mai posarem les dues variables amb el mateix valor*/

#define LCD 0
#define SERIAL_PORT 1

#define a1mux  8 //Canal selecció del multiplexor A1
#define a0mux  9 //Canal selecció del multiplexor A0
#define actmux 5 //Pin activació multiplexors
#define SENYAL 10 //Sortida senyal PWM
#define NUM_MITJANA 5 //Número de mostres per fer la mitjana mòvil
#define NUM_VOLTES 3 //Número de mostres per fer la mediana 

//Distància de separació entre els transductors NORD-SUD i EST-OEST
#define DIST_NS 0.245 
#define DIST_EO 0.245
#define LLINDAR_VEL_MIN 0.05 // Llindar velocitat mínima a mostrar
//Valors en hexadecimal del OCR1A pel PWM=399 (equivalent a 25 us)
#define OCR1AH_PWM 0x01
#define OCR1AL_PWM 0x8F
//Valors en hexadecimal del OCR1B pel PWM=199 (equivalent a 12,50 us)
#define OCR1BH_PWM 0x00
#define OCR1BL_PWM 0xC7

//OCR1A=26000 valor suficient gran per saturar el comparador i poder reiniciar-lo
#define OCR1AH_LONG 0x65
#define OCR1AL_LONG 0x90
//OCR1B=3200(equivalent a 200 us), temps entre la desactivació del PWM i captura d’entrada
#define OCR1BH_LONG 0x0C
#define OCR1BL_LONG 0x80

volatile unsigned int compt_pols; // Comptador dels polsos PWM
volatile unsigned int compt_voltes; //Comptador número de voltes
volatile unsigned int compt_mitjana; // Comptador número de mostres mitjana
volatile unsigned int calcula; //Avisa que hem fet totes les voltes
volatile unsigned int mux; //Indica en quin estat ens trobem per multiplexar
volatile unsigned int envia; //Podem enviar les dades per port sèrie o LCD
volatile unsigned int z;
volatile float toffinal[4];//Valor final del temps de vol
volatile float tof[4][NUM_VOLTES];//Temps de vol per fer la mediana
volatile float velocitat_ns[NUM_MITJANA];//Velocitat eix NS abans de fer la mitjana
volatile float velocitat_eo[NUM_MITJANA];//Velocitat eix EO abans de fer la mitjana

void setup()
{
  compt_pols = 0;
  pinMode(a1mux, OUTPUT);
  pinMode(a0mux, OUTPUT);
  pinMode(SENYAL, OUTPUT); 
  pinMode(actmux, OUTPUT);
  digitalWrite(actmux, HIGH);//multiplexors sempre a ON
  lcd.begin(16, 2);

  Serial.begin(9600);

  /*
   TCCR1A – Timer/Counter1 Control Register A
    Bit 7:6 – COM1A1:0: Compare Output Mode pel Canal A (Pin 15, Arduino Pin 9)
    Bit 5:4 – COM1B1:0: Compare Output Mode pel Canal B (Pin 16, Arduino Pin 10)
    Bit 3:2 - Reservat
    Bit 1:0 – WGM11:0: Waveform Generation Mode
   */
  TCCR1A = 0b00100011;//Canal B, Mode "Fast PWM" i "Clear Timer on Compare Match"

    /*
  TCCR1B – Timer/Counter1 Control Register B
    Bit 7 – ICNC1: Input Capture Noise Canceler
    Bit 6 – ICES1: Input Capture Edge Select
    Bit 5 – Reservat
    Bit 4:3 – WGM13:2: Waveform Generation Mode
    Bit 2:0 – CS12:0: Clock Select
    */
  TCCR1B = 0b11011001;//Cancel•lació de soroll, Flanc de pujada, Rellotge=16MHz, Mode "Fast PWM" 

    /*
  TIMSK1 – Timer/Counter1 Interrupt Mask Register
    Bit 7,6 – Reservat
    Bit 5 – ICIE1: Timer/Counter1, Input Capture Interrupt Enable
    Bit 4,3 – Reservat
    Bit 2 – OCIE1B: Timer/Counter1, Output Compare B Match Interrupt Enable
    Bit 1 – OCIE1A: Timer/Counter1, Output Compare A Match Interrupt Enable
    Bit 0 – TOIE1: Timer/Counter1, Overflow Interrupt Enable
    */
  TIMSK1 = 0b00000100;

   /*
  ADCSRB – ADC Control and Status Register B
    Bit 6 – ACME: Analog Comparator Multiplexer Enable
   */
ADCSRB = 0b00000000;//Multiplexor en OFF, s’utilitza pin predeterminat del comparador

  /*
   ACSR – Analog Comparator Control and Status Register
    Bit 7 – ACD: Analog Comparator Disable
    Bit 6 – ACBG: Analog Comparator Bandgap Select
    Bit 5 – ACO: Analog Comparator Output
    Bit 4 – ACI: Analog Comparator Interrupt Flag
    Bit 3 – ACIE: Analog Comparator Interrupt Enable
    Bit 2 – ACIC: Analog Comparator Input Capture Enable
    Bits 1, 0 – ACIS1, ACIS0: Analog Comparator Interrupt Mode Select
   */
  ACSR = 0b00000100;//Captura d'entrada del comparador activat, mode d'intrrupció: sortida de commutació

  //Posem els valors corresponents a un senyal de 40kHz en els OCR
  OCR1AH = OCR1AH_PWM;
  OCR1AL = OCR1AL_PWM;
  OCR1BH = OCR1BH_PWM;
  OCR1BL = OCR1BL_PWM;
}
// Servei Rutina d'Interrupció pel "Output Comapre Match" del registre B
ISR(TIMER1_COMPB_vect)
{
  ++compt_pols;
  switch (compt_pols)
  {
//Si hem enviat els 5 polsos desactivem la sortida PWM i canviem els registres OCR1
    case (5):
      {
        TCCR1A = 0b00000011;
        OCR1AH = OCR1AH_LONG;
        OCR1AL = OCR1AL_LONG;
        OCR1BH = OCR1BH_LONG;
        OCR1BL = OCR1BL_LONG;
        break;
      }

    case (6):
      {
        /*
  TIFR – Timer/Counter1 Interrupt Flag Register
    Bit 7,6 – Reservats
    Bit 5 – ICF1: Timer/Counter1, Input Capture Flag
    Bit 4,3 – Reservats
    Bit 2 – OCF1B:Timer/Counter1, Output Compare B Match Flag
    Bit 1 – OCF1A:Timer/Counter1, Output Compare A Match Flag
    Bits 0 – TOV1: Timer/Counter1, Overflow Flag
        */
        TIFR1 = 0b00100000; //Resetegem la bandera de captura d'entrada (ICF)
        TIMSK1 = 0b00100100;//Activem la interrupció per captures d'entrada(ICIE)

        break;
      }
//Control dels multiplexors i del número de mostres fetes 
    case (7):
      {
        ++mux;
        mux &= 0b11;
        digitalWrite(a1mux, mux & 0b10); // Bit 1 de mux
        digitalWrite(a0mux, mux & 0b01); // Bit 0 de mux

        if (mux == 0b00) // Si hem fet un volta entra dins de la condició
        {
          ++compt_voltes;
          if (compt_voltes == NUM_VOLTES)
          {
            compt_voltes = 0;
            ++compt_mitjana;
            calcula=1;//Enviem dades a calcular, hem fet totes les voltes
            
            if (compt_mitjana == NUM_MITJANA)
            {
             envia=1;
//Podem mostrar les dades perquè ja tenim les dades suficients per fer la mitjana
             compt_mitjana=0;
            }
          }
        }
        break;
      }
      
//Sortida PWM a on, reiniciem el comptador i posem els valors del senyal de 40kHz 
    case (8):
      {
        OCR1AH = OCR1AH_PWM;
        OCR1AL = OCR1AL_PWM;
        OCR1BH = OCR1BH_PWM;
        OCR1BL = OCR1BL_PWM;
        TCCR1A = 0b00100011;
        compt_pols = 0;
        break;
      }
  }
}

// Servei Rutina d'Interrupció per la captura d'entrada
ISR(TIMER1_CAPT_vect)
{
  //Lectura del valor del comptador a través dels registres ICR1L i ICR1H
  uint16_t valor_timer;
  uint8_t timer_low;
  timer_low = ICR1L; 
  valor_timer = ICR1H;
  valor_timer <<= 8;
  //Afegim els últims 8 bits (ICR1L) del comptador a la variable final
  valor_timer |= timer_low; 
  mux &= 0b11;//Passem valor dels dos últims bits de mux a binari
  tof[mux][compt_voltes] = valor_timer;
  TIMSK1 = 0b00000100; //Desactivem la captura d'interrupcions d'entrada
}

//Comparem la velocitat final amb el llindar de velocitat mínima que hem prefixat
float velocitat_corregida(float vel_final)
{
  float vel_final_abs = abs(vel_final);
  if (vel_final_abs == 0) 
  {
    return vel_final_abs;
  }
  if (vel_final_abs < LLINDAR_VEL_MIN)
  {
    vel_final = 0;
    return vel_final;
  }
  if (vel_final_abs > LLINDAR_VEL_MIN)
  {
    return vel_final;
  }
}

//Ordenem els números de l'array i agafem el valor intermedi, extracció mediana
float valor_intermedi( volatile float tof[4][NUM_VOLTES], int z)
{
  int i, j, flag = 1;
  int temp;
  float tof_intermedi;
  for (i = 1; (i <= NUM_VOLTES) && flag; i++)
  {
    flag = 0;
    for (j = 0; j < (NUM_VOLTES - 1); j++)
    {
      if (tof[z][j + 1] < tof[z][j])
      {
        temp = tof[z][j];
        tof[z][j] = tof[z][j + 1];
        tof[z][j + 1] = temp;
        flag = 1;
      }
    }
  }
  tof_intermedi=tof[z][1]*62.5e-9;
  return tof_intermedi;
}

void loop()
{
  if (calcula) 
  {
    for (z = 0; z <= 3; z++)
    {
      toffinal[z] = valor_intermedi(tof,z);
    }
    z = 0;
//Càlcul de les velocitats en els dos eixos
    float vel_ns = ((1 / (toffinal[1])) - (1 / (toffinal[0]))) * (DIST_NS / 2);
    float vel_eo = ((1 / (toffinal[3])) - (1 / (toffinal[2]))) * (DIST_EO / 2);
// Guardem les velocitats en un array que ens servirà per fer la mitjana
    velocitat_ns[compt_mitjana] = vel_ns;
    velocitat_eo[compt_mitjana] = vel_eo;

// Un cop s'ha fet la mediana i aplicat el llindar mínim fem la mitjana de 5 dades i les enviem per port sèrie o per la pantalla
   if (envia)
    {
//Calculem la mitjana de la velocitat en cada eix, en aquest cas per cada 5 mostres
float velocitatfinal_ns = velocitat_corregida((velocitat_ns[0]+velocitat_ns[1]+velocitat_ns[2]+velocitat_ns[3]+velocitat_ns[4])/NUM_MITJANA);

float velocitatfinal_eo = velocitat_corregida((velocitat_eo[0]+velocitat_eo[1]+velocitat_eo[2]+velocitat_eo[3]+velocitat_eo[4])/NUM_MITJANA);

//Calculem la direcció del vent respecte l’eix EO
float direccio = (atan2(velocitatfinal_ns, velocitatfinal_eo)) * (57.2958); 

    if (SERIAL_PORT==1)//Envia dades pel port sèrie
      {
        Serial.print("NS: ");
        Serial.println(velocitatfinal_ns);
        Serial.print("                EO: ");
        Serial.println(velocitatfinal_eo);
        Serial.print("                                   DIR: ");
        Serial.println(direccio);
        envia=0;
      }
    if (LCD==1)//Envia dades per la pantalla LCD
      {
       lcd.setCursor(0, 0);
       lcd.print("NS");
       lcd.setCursor(3, 0);
       lcd.print(velocitatfinal_ns,2);
       lcd.setCursor(0, 1);
       lcd.print("EO");
       lcd.setCursor(3, 1);
       lcd.print(velocitatfinal_eo,2);
       lcd.setCursor(10, 1);
       lcd.print(direccio,2);
      }
      calcula = 0; //Posem calcula a 0 fins que es torni a fer tot el càlcul de nou
      }
    }
  }

